//
//							16-bit Arithmetic Logic Unit
//
//					(C) Copyright 2007-2008 John B. Stephensen
//
// This Verilog source file and all its derivatives are licensed only for
// personal non-profit educational use in the Amateur Radio Service and
// the license is not transferrable. The information is provided as-is for
// experimental purposes and the author does not warranty its freedom
// from defects or its suitability for any specific application.
//
// The ALU has 8 inputs:
//		A - Data from read/write port of register file
//		B - Data from read-only port of register file
//		C - 8-bit constant value from insturuction field
//		D0 - Data from I/O ports (source determined externally)
//		D1 - Data from memory (source determined externally)
//		D7 - Data from instruction (source determined externally)
//		OP - operation type (output multiplexer selecting logic block)
//		MOD - operation modifier (input to logic block)
//
// The IV input signals that data is valid. The instruction to be executed is encoded in
// the OP and MOD inputs. OP selects a functional unit to carry out the instruction by
// selecting the output of 1 of 8 units. MOD specifies the type of operation to be carried
// out in the selected functional unit. The following functions are provided:
//
//		1) 16-bit arithmetic with register A and register B
//		2) Increment and decrement register A
//		3) Complex 8x8-bit arithmetic with register A and register B 
//		3)	16-bit logical operations on register A and register B
//		4) Single bit operations on bits in register A
//		5) Shift register A right or left by 1 bit
//		6) Copy 8 bit field from register B to register A and optionally extend sign
//		7) 16x16-bit multiply and 32-bit accumulate
//		8) 16-bit integer and fractional division
//		9) Load data from memory, I/O port or instruction field
//
//	The ALU performs the following arithmetic and logical operations:
//
//		op		mod	Inst.	Comments
//		---	---	-----	-------------------------------------------
//		000	xxx	I/O	F = D0
//		001	xxx	R/W	F = D1
//		010	000	MOV	F = B
//		010	001	NEG	F = -B
//		010	010	ADD	F = A + B
//		010	011	SUB	F = A - B
//		010	100	ADC	F = A + B + CI
//		010	101	SBC	F = A + ~B + CI
//		010	110	ADI	F = A + C
//		010	111	SUI	F = A - C
//		011	000	MOV	Fx = Ax + 1, Fy = Ay
//		011	011	CNEG	Fx = Ax - 1, Fy = Ay
//		011	100	CADD	Fx = Ax + Bx, Fy = Ay + By 
//		011	111	CSUB	Fx = Ax - Bx, Fy = Ay - By 
//		100	000	NOT	F = ~B
//		100	001	AND	F = A and ~B
//		100	010	OR		F = A or B
//		100	011	XOR	F = A xor B
//		100	100	MASK	F[C:0] = A[C:0], F[15:C+1] = 0
//		100	101	CLR	F[C] = 0
//		100	110	SET	F[C] = 1
//		100	111	INV	F[C] = ~A[C]
//		101	000	SXL	F = 8xB[7],B[7:0]
//		101	001	SXH	F = 8xB[15],B[15:8]
//		101	010	RR4	F = B[3:0],A[15:4]
//		101	011	RR8	F = B[7:0],A[15:8]
//		101	100	RR12	F = B[11:0],A[15:12]
//		101	101	REV	F = B[0:15]
//		101	110	SHL	F = A << 1 (C provides modifier bits)
//		101	111	SHR	F = A >> 1 (C provides modifier bits)
//		110	00x	LPL	F = P[15:0]
//		110	01x	LPH	F = P[31:16]
//		110	10x	LQ		F = Q
//		110	11x	LR		F = R
//		111	xxx	LDx	F = D7
//
// Compares, bit tests, multiplies and divides don't use the output multiplexer and are initiated
// by CMP, TST, MV or DV:
//
//		op		mod	Inst.	Comments
//		---	---	-----	------------------------
//		100	S00	MUL	P = A * B
//		100	S01	MAC	P = P + (A * B)
//		100	S10	MULN	P = -(A * B)
//		100	S11	MSUB	P = P - (A * B)
//		101	0x0	FDIV	Q = rem/B
//		101	0x1	IDIV	Q = A div B, R = A rem B
//		101	100	TST	link = A[C]
//		101	101	SETL	link = C[0]
//		101	11x	SETC	carry = C[0]
//		110	011	CMP	A-B
//		110	101	CMPC	A-B + carry
//
// Multiplication may use signed or unsigned operands. The S bit selects signed operations when 1.
//
// The result of these operations appears on the F output. The other outputs are the zero, carry
// and link flags.
//
// With multipler and divider: 213 slices used (4% XC3S500E). 160 MHz maximum clock rate.
// No multipler or divider: 135 slices used (2% XC3S500E). 160 MHz maximum clock rate.
// Maximum combinitorial logic delay is in 16-bit adder/subtracter.
//
// Normal warnings:
//		Signals <prod<35:34>> and <prod<1:0>> are assigned but never used.
//
// History:
//		M001	4-16-08	Reduced propogation delay by adding duplicate arithmetic and logic units.
//		M002	6-8-08	Removed convolutional encoder (more efficient as I/O device)
//		M003	8-5-08	Moved output multiplexer ports to allow 16-bit instruction format
//		M004	8-9-08	32-bit unsigned divider logic replaces 16-bit fractional divider
//		M005	8-12-08	Removed barrel shifer and replaced with byte swap and bit shifter (-112 slices)
//		M006	9-29-08	Allow comparison and bit test operations by inhibiting storage in register
//		M007	3-24-09	Added mask instruction
//		M008	3-28-08	Add set/reset carry and link bit instructions - move w/o op decode to this module
//
module ALU16(a, b, c, d0, d1, d7, iv, wov, op, mod, f, zero, carry, link, clk, rst);
    input [15:0] a,b;			// data from registers A and B
	 input [3:0] c;				// B address field
	 input [15:0] d0,d1,d7;		// data from I/O ports, memory and constants in instructions
	 input iv,wov;					// inputs valid for ALU read/write and write-only instructions
    input [2:0] op;				// select functional unit to perform operation
	 input [2:0] mod;				// operation modifier
    output [15:0] f;				// result
	 output zero,carry,link;	// zero, carry and link flags for math and shift operations
	 input clk,rst;				// clock and reset for carry bit and multiplier
// multiplier/divider generation control
`define muldiv
// internal signals
wire [15:0] bmux;	// B-input multiplexer
wire cp;				// carry propagate
wire en;				// latch carry and zero bits
reg ci;				// carry flag for adder/subtractor
wire co;				// carry out bit adder/subtractor
wire z16;			// zero detector output
reg z;				// arithmetic zero flag
wire [3:0] bitsel;// bit select
wire bitmux;		// selected bit
reg linkbit;		// link bit
wire sv;				// shift enable
wire si;				// shift end bit
wire [15:0] f2,f3,f4,f5,f6;	// ALU functional unit outputs
wire cmp,tst,mv,dv,csr;			// write-only operations
// decode write-only operations
assign mv = wov & (op == 4);							// multiply operations
assign dv = wov & (op == 5) & (mod[2:1] == 0);	// divide operations
assign tst = wov & (op == 5) & (mod[2:1] == 2);	// link bit operations
assign csr = wov & (op == 5) & (mod[2:1] == 3);	// carry bit operations
assign cmp = wov & (op == 6);							// comparison operations
// 16-bit adder or subtractor using registers A and B or B only
// 16-bit incrementer/decrementer using register A and constant C
assign bmux = (mod[2:1] == 3) ? {12'h000,c} : b;	// select constant when mod = 6 or 7
assign cp = (mod[2:1] == 2);								// enable carry prop. when mod = 4 or 5
addsubmovneg16a au16 (
	.A(a),						// accumulator register
	.B(bmux),					// another register or a constant
	.CI(cp ? ci : mod[0]),	// carry input is zero or previous output
	.CO(co),						// carry output
	.SUB(mod[0]),				// select add (0) or subtract (1)
	.ENA(mod[2]|mod[1]),		// select zero or A input (zero when mod = 0 or 1)
	.Y(f2)						// result
	);
// detect zero output
zero16 zerodet (
	.A(f2),
	.Y(z16)
	);
// latch carry out bit, detect overflow and check for zero
assign en = cmp | (iv & (op == 2));
always @ (posedge clk)
begin
	if (rst) ci <= 1'b0;
	else if (en|csr) ci <= csr ? c[0] : co;	// last carry out
	if (rst) z <= 1'b0;
	else if (en) z <= z16;
end
assign carry = ci;
assign zero = z;
// Dual 8-bit adder or subtractor using registers A and B
// 8-bit incrementer/decrementer using register A
addsubmovneg8a au8l (
	.A(a[7:0]),		// accumulator
	.B(b[7:0]),		// another register
	.CI(mod[1] & mod[0]),
	.CO(),
	.SUB(mod[1] & mod[0]),
	.ENA(mod[1]),
	.Y(f3[7:0])
	);
addsubmovneg8a au8h (
	.A(a[15:8]),	// accumulator
	.B(b[15:8]),	// another register
	.CI(mod[2] & mod[0]),
	.CO(),
	.SUB(mod[2] & mod[0]),
	.ENA(mod[2]),
	.Y(f3[15:8])
	);
// logical operations on one bit in register A - bit select, reset, set or toggle
lu16a lu (
	.a(a),
	.b(b),
	.c(c),
	.op(mod),
	.y(f4)
	);
// Select bits to test or append to shifted word
assign sv = tst | (iv & (op == 5) & (mod[2:1] == 2'b11)); // sample when comparing or shifting
assign bitsel = tst ? c : {c[0],c[0],c[0]};	// select bit 0 or bit 15 when shifting
MUX16 smux1 (
	.D(a),		// sample reg. A bits
	.S(bitsel),	// select bit
	.Y(bitmux)	// output
	);
// link bit is set/reset by bit multiplexer data when shifting (op=5) or testing bits (tst & mod[0]=0)
// link set/reset instruction (tst & mod[0]=1) uses c[0] as the input
always @ (posedge clk)
begin
	if (rst) linkbit <= 0;
	else if (sv) linkbit <= (tst & mod[0]) ? c[0] : bitmux;
end
assign link = linkbit;
// shift logic for end bits
MUX4 smux16 (
	.D({c[1],linkbit,a[15],a[0]}),	// select 0, 1, link bit, bit 15 or bit 0
	.S(c[3:2]),								// inst. bits 7-6 select shifter input
	.Y(si)
	);
// shift multiplexer
MUX8X16 dmux (
	.D0({b[7],b[7],b[7],b[7],b[7],b[7],b[7],b[7],b[7:0]}),	// extend sign of lower byte
	.D1({b[15],b[15],b[15],b[15],b[15],b[15],b[15],b[15],b[15:8]}),	// extend sign of upper byte
	.D2({b[3:0],a[15:4]}),		// rotate right and insert 4 bits
	.D3({b[7:0],a[15:8]}),		// rotate right and insert 8 bits
	.D4({b[11:0],a[15:12]}),	// rotate right and insert 12 bits
	.D5({b[0],b[1],b[2],b[3],b[4],b[5],b[6],b[7],b[8],b[9],b[10],b[11],b[12],b[13],b[14],b[15]}), // reverse bits
	.D6({si,a[15:1]}),			// shift right by 1 bit
	.D7({a[14:0],si}),			// shift left by 1 bit
	.S(mod[2:0]),
	.Y(f5)
	);
// output multiplexer - use optimized multiplxers
MUX8X16 fmux (
	.D0(d0),		// I/O ports
	.D1(d1),		// memory
	.D2(f2),		// scalar arithmetic
	.D3(f3),		// complex arithmetic
	.D4(f4),		// logic
	.D5(f5),		// shift
	.D6(f6),		// multiply/divide
	.D7(d7),		// memory
	.S(op),
	.Y(f)
	);
// select whether multiplier and divider are generated
`ifdef muldiv
// internal signals
wire [31:0] sum;	// sum of products
wire [15:0] q,r;	// quotient and remainder from divider
// multiplier/accumulator
wire [35:0] prod;	// full product
reg mv0;				// latch sum of products
reg mod0,mod1;		// delay operation modifier
// multiplier - 1 clock delay
MULT18X18SIO #(
	.AREG(0),		// enable output register only
	.BREG(0),
	.PREG(1)
	) mult (
	.A({mod[2] & a[15],a,1'b0}),	// trim inputs to 16 bits and
	.B({mod[2] & b[15],b,1'b0}),	// select signed or unsigned inputs
	.P(prod),
	.CLK(clk),
	.CEP(mv),		// save product
	.RSTP(1'b0)		// never reset
	);
// delay data valid and modifier by 1 clock to match product
always @ (posedge clk)
begin
	mv0 <= mv;
	mod0 <= mod[0];
	mod1 <= mod[1];
end
// accumulator - no delay on output - data registered internally
// load accumulator during multiplication if MUL or MULN instruction
// add or subtract during MAC or MSUB instruction
acc32a acc (
	.B(prod[33:2]),
	.SUB(mod0),
	.ACC(mod1),
	.Y(sum),
	.CE(mv0),
	.CLK(clk),
	.SCLR(1'b0)
	);
// 16-bit divider
div16 div (
	.ain(a),			// numerator and denominator input
	.bin(b),
	.iv(dv),			// load data and start
	.mod(mod[0]),	// 1=division with remainder, 0=generate fractional remainder
	.qout(q),		// quotient
	.rout(r),		// remainder
	.clk(clk),
	.rst(rst)
	);
// Retreive product, quotient and remainder 
MUX4X16 pqrmux (
	.D0(sum[15:0]), 
	.D1(sum[31:16]), 
	.D2(q), 
	.D3(r), 
	.S(mod[1:0]), 
	.Y(f6)
	);
`else
// No multipler or divider
assign f6 = 16'hFFFF;
`endif
endmodule
