//
//				div16.v - Divide unsigned number by unsigned number
//
// 				(C) Copyright 2006-2008 John B. Stephensen
//
// This Verilog source file and all its derivatives are licensed only for
// personal non-profit educational use in the Amateur Radio Service and
// the license is not transferrable. The information is provided as-is for
// experimental purposes and the author does not warranty its freedom
// from defects or its suitability for any specific application.
//
// The dividend and divisor are loaded when IV and MOD are true. On each
// successive clock after IV is asserted, the divisor (B) is subtracted
// from the upper half of the dividend shift register (A) if the result
// will be positive. The result is shifted left by one bit and a one
// shifted into the quotient register (Q) if the subtraction succeeded.
// When the calculation is complete, Q contains the integer quotient
// and the most significant 16 bits of A contain the integer remainder.
// If IV is asserted with MOD low, the calculation continuues and 16
// fractional quotient bits are generated in Q. The upper bit of ROUT
// is one while calulations are proceeding and returns to zero when
// they are complete.
//
// 52 slices used. 165 MHz maximum clock speed. 17 or 18 clock delay.
//
// Normal Warnings:
//		none
//
module div16(ain, bin, mod, iv, qout, rout, ov, clk, rst);
    input [15:0] ain,bin;	// dividend and divisor input
	 input mod;					// 1=integer remainder, 0=fractional remainder
    input iv;					// inputs valid
    output [15:0] qout;		// quotient
    output [15:0] rout;		// remainder
    output ov;					// quotient and remainder valid
	 input clk;					// master clock
	 input rst;					// master reset
// Internal signals
wire[15:0] d;			// dividend less divisor
wire neg;				// true if subtraction produced negative result
reg [31:0] a;			// dividend/remainder shift register
reg [15:0] b;			// divisor register
reg [15:0] q;			// quotient shift register
reg [5:0] c;			// bit counter
wire nz;					// bit counte not zero
// subtract divisor
assign d = a[31:16] - b;
assign neg = d[15];
assign nz = |c;
always @ (posedge clk)
begin
	// bit counter 
	if (rst) c <= 0;
	else if (iv) c <= {4'b1000,mod};	// load with bits + 1
	else if (nz) c <= c - 1;			// then count down to 0
	// load divisor register
	if (iv & mod) b <= bin;
	// load dividend or shift remainder
	if (iv & mod) a <= {16'h0000,ain};
	else if (nz) a <= {(neg ? a[30:16] : d[14:0]),a[15:0],1'b0};
	// accumulate quotient in shift register
	if (iv) q <= 32'h00000000;
	else if (nz) q <= {q[14:0],~neg};
end
// connect output
assign qout = q;					// quotient
assign rout = {nz,a[31:17]};	// remainder in upper half of A
assign ov = ~nz;
endmodule
